@@ -13,9 +13,10 @@ class BaseDataProcessor(object):
         self.chemistry_feature = read_feature_table('chemistry', self.csv_dir)
         self.coagulation_feature = read_feature_table('coagulation', self.csv_dir)
         self.blood_feature = read_feature_table('complete_blood_count', self.csv_dir)
-        self.vitalsign_feature = read_feature_table('vitalsign', self.csv_dir)
         self.weight_feature = read_feature_table('weight', self.csv_dir)
         self.height_feature = read_feature_table('height', self.csv_dir)
+        self.vitalsign_feature = read_feature_table('vitalsign', self.csv_dir)
+        self.vitalsign_feature = self.vitalsign_feature.drop_duplicates().reset_index(drop=True)
 
     '''
         @param: {start, stop} two integer represent how long a time window is
@@ -139,6 +140,47 @@ class BaseDataProcessor(object):
         res = cohort_with_time.merge(feature_df, how='left', on=['subject_id', 'episode_id'])
         res = res.drop('charttime', axis=1)
         return res
+    
+
+    """
+        functionality similar to {merge_feature_to_time_dataframe}, but the size of vitalsign dataframe is huge
+        It has to be sliced into several chunks and process separately, otherwise may exceed memory limit
+    """
+    def merge_vitalsign_to_time_dataframe(self, cohort_with_time:pd.DataFrame, vitalsign_feature:pd.DataFrame, cohort_df:pd.DataFrame, agg:str, time_step:int):
+        feature_df = cohort_df.merge(vitalsign_feature, how='inner', on='subject_id')
+        # first filter the feature events that happens outside the ICU stay time
+        feature_df = feature_df.loc[(feature_df['charttime'] >= feature_df['starttime']) & (feature_df['charttime'] <= feature_df['endtime'])]
+        feature_df = feature_df.sort_values(by='charttime').reset_index()
+        feature_df = feature_df.drop(['starttime', 'endtime', 'index'], axis=1)
+
+        # determine the number of rows, and horizontally slice the dataframe into several chunks
+        # process each chunk independently to avoid exceeding memory limit
+        process_list = []
+        length = feature_df.shape[0]
+        num_chunks = 30
+        interval = int(length * (1/num_chunks))
+
+        # slice the dataframe into chunks
+        cur = 0
+        for i in range(1, num_chunks):
+            process_list.append(feature_df.loc[cur : cur + interval])
+            cur += interval
+        process_list.append(feature_df.loc[cur : length])
+
+        for i in range(len(process_list)):
+            res = cohort_with_time.merge(process_list[i], how='inner', on=['subject_id', 'episode_id'])
+            # match each event to its closest time slot
+            res = res.loc[(res['charttime'] >= res['time']) & (res['charttime'] <= res['time'] + pd.Timedelta(time_step, 'h'))]
+            res = res.groupby(by=['episode_id', 'time']).agg(agg)
+            process_list[i] = res
+        
+        # concat the chuncks back into a single dataframe
+        res = pd.concat(process_list)
+        res = cohort_with_time.merge(res, how='left', on=['subject_id', 'time'])
+        # regroup to remove multiple matches
+        res = res.groupby(by=['time', 'subject_id', 'episode_id']).agg(agg)
+        res = res.sort_values(['subject_id', 'episode_id', 'time']).reset_index().drop('charttime', axis=1)
+        return res
 
 
     '''
@@ -196,7 +238,7 @@ class VancomycinDosingDataProcessor(BaseDataProcessor):
         # use the glucose from the vitalsign table instead of chemistry
         self.chemistry_feature = self.chemistry_feature.drop('glucose', axis=1)
         # merge all features together with task_df
-        cohort_with_time = self.create_time_cohort(self.cohort_df, time_step)
+        self.cohort_with_time = self.create_time_cohort(self.cohort_df, time_step)
 
         feature_table = {
             'chemistry':self.chemistry_feature, 
@@ -207,24 +249,21 @@ class VancomycinDosingDataProcessor(BaseDataProcessor):
             'weight':self.weight_feature
             }
         
-        processed_list = []
+        action_reward_cohort = self.merge_action_to_time_dataframe(self.cohort_with_time, self.task_df, agg, time_step)
 
         for name, feature in feature_table.items():
             if name == 'height':
-                temp = self.merge_height_to_time_dataframe(cohort_with_time, feature, self.cohort_df, agg)
+                temp = self.merge_height_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg)
             elif name == 'weight':
-                temp = self.merge_weight_to_time_dataframe(cohort_with_time, feature, self.cohort_df, agg)
+                temp = self.merge_weight_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg)
+            elif name == 'vitalsign':
+                temp = self.merge_vitalsign_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg, time_step)
             else:
-                temp = self.merge_feature_to_time_dataframe(cohort_with_time, feature, self.cohort_df, agg, time_step)
-            processed_list.append(temp)
-
-        action_reward_cohort = self.merge_action_to_time_dataframe(cohort_with_time, self.task_df, agg, time_step)
-        processed_list.append(action_reward_cohort)
-
-        for i in range(1, len(processed_list)):
-            processed_list[0] = processed_list[0].merge(processed_list[i], how='left', on=['subject_id', 'episode_id', 'time'])
-
-        self.featured_cohort_with_time = processed_list[0]
+                temp = self.merge_feature_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg, time_step)
+            action_reward_cohort = action_reward_cohort.merge(temp, how='left', on=['subject_id', 'episode_id', 'time'])
+            del temp
+        
+        self.featured_cohort_with_time = action_reward_cohort
     
 
     '''
@@ -280,36 +319,31 @@ class HeparinDosingDataProcessor(BaseDataProcessor):
         # use the glucose from the vitalsign table instead of chemistry
         self.chemistry_feature = self.chemistry_feature.drop('glucose', axis=1)
         # merge all features together with task_df
-        cohort_with_time = self.create_time_cohort(self.cohort_df, time_step)
+        self.cohort_with_time = self.create_time_cohort(self.cohort_df, time_step)
 
         feature_table = {
             'chemistry':self.chemistry_feature, 
-            # 'coagulation':self.coagulation_feature, 
             'vitalsign':self.vitalsign_feature,
             'blood':self.blood_feature, 
             'height':self.height_feature, 
             'weight':self.weight_feature
             }
 
-        processed_list = []
+        action_reward_cohort = self.merge_action_to_time_dataframe(self.cohort_with_time, self.task_df, agg, time_step)
 
         for name, feature in feature_table.items():
             if name == 'height':
-                temp = self.merge_height_to_time_dataframe(cohort_with_time, feature, self.cohort_df, agg)
+                temp = self.merge_height_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg)
             elif name == 'weight':
-                temp = self.merge_weight_to_time_dataframe(cohort_with_time, feature, self.cohort_df, agg)
+                temp = self.merge_weight_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg)
+            elif name == 'vitalsign':
+                temp = self.merge_vitalsign_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg, time_step)
             else:
-                temp = self.merge_feature_to_time_dataframe(cohort_with_time, feature, self.cohort_df, agg, time_step)
-            processed_list.append(temp)
-
-        action_reward_cohort = self.merge_action_to_time_dataframe(cohort_with_time, self.task_df, agg, time_step)
-        processed_list.append(action_reward_cohort)
-        self.temp = processed_list.copy()
-
-        for i in range(1, len(processed_list)):
-            processed_list[0] = processed_list[0].merge(processed_list[i], how='left', on=['subject_id', 'episode_id', 'time'])
-
-        self.featured_cohort_with_time = processed_list[0]
+                temp = self.merge_feature_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg, time_step)
+            action_reward_cohort = action_reward_cohort.merge(temp, how='left', on=['subject_id', 'episode_id', 'time'])
+            del temp
+        
+        self.featured_cohort_with_time = action_reward_cohort
 
 
     def save_task_df(self, csv_dir:Path, filename:str):
